﻿using Infragistics;
using Infragistics.Windows;
using Infragistics.Windows.DataPresenter;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Collections.Specialized;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Data;
using System.Windows.Documents;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Navigation;
using System.Windows.Shapes;

namespace DataGridUndoDelete
{
	/// <summary>
	/// Interaction logic for MainWindow.xaml
	/// </summary>
	public partial class MainWindow : Window
	{
		private ObservableCollection<Person> _people;

		public MainWindow()
		{
			_people = new ObservableCollection<Person>();
			_people.Add(new Person { FirstName = "John", LastName = "Doe", Dob = new DateTime(1970, 1, 1) });

			this.DataContext = _people;

			InitializeComponent();
		}

		private void grid_RecordsDeleting(object sender, Infragistics.Windows.DataPresenter.Events.RecordsDeletingEventArgs e)
		{
			object firstSourceItems = e.Records[0].RecordManager.SourceItems;
			e.UndeleteStrategy = new ListUndeleteStrategy(e.Records);
		}
	}

	public class Person : PropertyChangeNotifierExtended
	{
		private string _firstName;
		public string FirstName
		{
			get { return _firstName; }
			set { this.SetField(ref _firstName, value, "FirstName"); }
		}

		private string _lastName;
		public string LastName
		{
			get { return _lastName; }
			set { this.SetField(ref _lastName, value, "LastName"); }
		}

		private DateTime _dob;
		public DateTime Dob
		{
			get { return _dob; }
			set { this.SetField(ref _dob, value, "Dob"); }
		}
	}

	#region UnboundCellRecordUndeleteStrategy
	/// <summary>
	/// Base UndeleteRecordsStrategy that provides storage of unbound cell values.
	/// </summary>
	public abstract class UnboundCellRecordUndeleteStrategy : UndeleteRecordsStrategy
	{
		#region Member Variables

		private List<UnboundField> _unboundFields;
		private Dictionary<object, IList<object>> _unboundValues;
		private Dictionary<object, object> _newToOldMapping;

		#endregion //Member Variables

		#region Constructor
		/// <summary>
		/// Initializes a new <see cref="RecordsWithUnboundCellUndeleteStrategy"/>
		/// </summary>
		/// <param name="records">The list of records being deleted</param>
		protected UnboundCellRecordUndeleteStrategy(IList<Record> records)
		{
			if (records == null)
				throw new ArgumentNullException("records");

			if (records.Count > 0)
			{
				List<UnboundField> unboundFields = new List<UnboundField>();
				FieldLayout fl = records[0].FieldLayout;

				foreach (Field field in fl.Fields)
				{
					UnboundField ubField = field as UnboundField;

					if (null == ubField)
						continue;

					if (ubField.BindingPath != null)
					{
						switch (ubField.BindingMode)
						{
							case BindingMode.OneWay:
							case BindingMode.Default:
							case BindingMode.OneTime:
								continue;
						}
					}

					unboundFields.Add(ubField);
				}

				int unboundFieldCount = unboundFields.Count;

				if (unboundFieldCount > 0)
				{
					_unboundFields = unboundFields;
					_unboundValues = new Dictionary<object, IList<object>>();

					foreach (Record record in records)
					{
						DataRecord dataRecord = record as DataRecord;

						if (null == dataRecord)
							continue;

						object dataItem = dataRecord.DataItem;

						if (dataItem == null)
							continue;

						object[] unboundValues = new object[unboundFieldCount];

						for (int i = 0; i < unboundFieldCount; i++)
							unboundValues[i] = dataRecord.Cells[unboundFields[i]].Value;

						_unboundValues[dataItem] = unboundValues;
					}
				}
			}
		}
		#endregion //Constructor

		#region Base class overrides

		#region Undelete
		/// <summary>
		/// Invoked to perform an undeletion of records associated with the specified data items.
		/// </summary>
		/// <param name="oldRecords">A list of objects containing information about the records that were deleted.</param>
		/// <returns>Returns a dictionary that provides a mapping from the old data items to the new data items.</returns>
		public override IDictionary<RecordInfo, object> Undelete(IList<RecordInfo> oldRecords)
		{
			IDictionary<RecordInfo, object> oldToNewMapping = this.UndeleteOverride(oldRecords);

			if (null != oldToNewMapping && oldToNewMapping.Count > 0)
			{
				_newToOldMapping = new Dictionary<object, object>();

				foreach (KeyValuePair<RecordInfo, object> pair in oldToNewMapping)
					_newToOldMapping[pair.Value] = pair.Key.DataItem;
			}

			return oldToNewMapping;
		}
		#endregion //Undelete

		#region ProcessUndeletedRecords
		/// <summary>
		/// Invoked after the call to <see cref="Undelete(IList{RecordInfo})"/> to allow additional processing of the new records.
		/// </summary>
		/// <param name="recordsCreated">A list of the <see cref="DataRecord"/> instances for the undeleted records based on the specified mapping.</param>
		public override void ProcessUndeletedRecords(IList<DataRecord> recordsCreated)
		{
			if (_newToOldMapping == null || _newToOldMapping.Count == 0)
				return;

			// we're only using this to reset the unbound cell values
			if (_unboundFields == null || _unboundFields.Count == 0)
				return;

			foreach (DataRecord newRecord in recordsCreated)
			{
				object oldDataItem;

				if (!_newToOldMapping.TryGetValue(newRecord.DataItem, out oldDataItem))
					continue;

				IList<object> unboundValues;

				if (!_unboundValues.TryGetValue(oldDataItem, out unboundValues))
					continue;

				for (int i = 0; i < unboundValues.Count; i++)
				{
					Field fld = _unboundFields[i];

					Debug.Assert(fld.Owner == newRecord.FieldLayout, "Record is associated with a different field layout!");

					// if somehow the field was removed then skip it
					if (fld.Owner == newRecord.FieldLayout && fld.Index >= 0)
						newRecord.Cells[fld].Value = unboundValues[i];
				}
			}
		}
		#endregion //ProcessUndeletedRecords

		#endregion //Base class overrides

		#region Methods

		#region UndeleteOverride

		/// <summary>
		/// Invoked to perform an undeletion of records associated with the specified data items.
		/// </summary>
		/// <param name="oldRecords">A list of objects containing information about the records that were deleted.</param>
		/// <returns>Returns a dictionary that provides a mapping from the old data items to the new data items.</returns>
		protected abstract IDictionary<RecordInfo, object> UndeleteOverride(IList<RecordInfo> oldRecords);

		#endregion //UndeleteOverride

		#endregion //Methods
	}
	#endregion //UnboundCellRecordUndeleteStrategy

	#region ListUndeleteStrategy
	/// <summary>
	/// Custom <see cref="UndeleteStrategy"/> for IBindingList and INotifyCollectionChanged data sources.
	/// </summary>
	public class ListUndeleteStrategy : UnboundCellRecordUndeleteStrategy
	{
		#region Constructor
		/// <summary>
		/// Initializes a new <see cref="ListUndeleteStrategy"/>
		/// </summary>
		/// <param name="records">The list of records being deleted</param>
		public ListUndeleteStrategy(IList<Record> records)
			: base(records)
		{
			// since this class is set up to reuse the same old items, we can only
			// support a subset of list types. also since we want to ensure the 
			// grid knows when the records are undeleted we're limiting this to 
			// known classes that send collection change notifications as well.
			foreach (Record record in records)
			{
				IEnumerable dataSource = record.RecordManager.SourceItems;

				if (dataSource is DataView || dataSource is DataViewManager)
					throw new InvalidOperationException("This class cannot be used with datatables/dataviews since it just reinserts the same item back into the collection.");

				if (dataSource is IBindingList)
					continue;

				if (dataSource is IList && dataSource is INotifyCollectionChanged)
					continue;

				throw new ArgumentException("The ListUndeleteStrategy is only supported with IBindingList and INotifyCollectionChanged sources");
			}
		}
		#endregion //Constructor

		#region Base class overrides

		#region CanUndelete
		/// <summary>
		/// Invoked to determine if an undeletion of the specified data items is possible.
		/// </summary>
		/// <param name="oldRecords">A list of objects containing information about the records that were deleted.</param>
		public override bool CanUndelete(IList<UndeleteRecordsStrategy.RecordInfo> oldRecords)
		{
			return true;
		}
		#endregion //CanUndelete

		#region UndeleteOverride
		/// <summary>
		/// Invoked to perform an undeletion of records associated with the specified data items.
		/// </summary>
		/// <param name="oldRecords">A list of objects containing information about the records that were deleted.</param>
		/// <returns>Returns a dictionary that provides a mapping from the old data items to the new data items.</returns>
		protected override IDictionary<RecordInfo, object> UndeleteOverride(IList<RecordInfo> oldRecords)
		{
			Dictionary<RecordInfo, object> oldToNewMapping = new Dictionary<RecordInfo, object>();

			RecordInfo[] records = oldRecords.ToArray();
			Comparison<RecordInfo> comparison = delegate(RecordInfo item1, RecordInfo item2)
			{
				return item1.DataItemIndex.CompareTo(item2.DataItemIndex);
			};

			// sort by the original index to ensure we get them in the original order
			Utilities.SortMergeGeneric(records, Utilities.CreateComparer(comparison));

			foreach (RecordInfo record in records)
			{
				RecordManager rm = record.RecordManager;
				IList list = rm.SourceItems as IList;

				if (list == null)
					continue;

				int newIndex = Math.Min(list.Count, record.DataItemIndex);
				object newDataItem = record.DataItem;

				list.Insert(newIndex, newDataItem);

				// since we're reusing the same object the old and new will be the same
				oldToNewMapping[record] = newDataItem;
			}

			return oldToNewMapping;
		}
		#endregion //UndeleteOverride

		#endregion //Base class overrides
	}
	#endregion //ListUndeleteStrategy

}
